/*
    Scanbot SDK Flutter Plugin

    Copyright (c) 2019 doo GmbH. All rights reserved.
 */
package io.scanbot.sdk.flutter

import android.util.Base64
import androidx.core.net.toUri
import com.google.gson.Gson
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import io.scanbot.sdk.flutter.camera.config.BarcodeScannerConfiguration
import io.scanbot.sdk.flutter.camera.config.ContourConfiguration
import io.scanbot.sdk.flutter.camera.config.CroppingViewConfig
import io.scanbot.sdk.flutter.camera.config.DocumentScannerConfiguration
import io.scanbot.sdk.flutter.camera.config.FinderConfig
import io.scanbot.sdk.flutter.camera.config.MedicalCertificateScannerConfiguration
import io.scanbot.sdk.flutter.camera.config.SelectionOverlayConfiguration
import io.scanbot.sdk.flutter.utils.JsonUtils
import io.scanbot.sdk.ocr.process.OcrResult
import io.scanbot.sdk.persistence.Page
import io.scanbot.sdk.persistence.PageStorageProcessor
import org.json.JSONObject
import java.io.File

fun ocrResultToJson(ocrResult: OcrResult, newFile: File?): JSONObject {
    val ocrResultJson = JsonUtils.resultToJson(ocrResult)

    ocrResult.sandwichedPdfDocumentFile?.toUri()?.toString()?.let {
        ocrResultJson.put(
            "pdfFileUri",
            it
        )
    }

    newFile?.toUri()?.toString()?.let {
        ocrResultJson.put(
            "pdfFileUri",
            it
        )
    }

    ocrResult.recognizedText?.let {
        ocrResultJson.put("plainText", it)
    }

    return ocrResultJson
}

typealias PagesJson = List<HashMap<String, Any>>
typealias PolygonJson = List<HashMap<String, Any>>
typealias Json = HashMap<String, Any>
typealias LanguagesJson = ArrayList<String>

fun Json.toPage(gson: Gson): Page {
    return gson.fromJson(JSONObject(this as Map<*, *>).toString(), Page::class.java)
}

fun Json.toFinderConfig(gson: Gson): FinderConfig {
    return gson.fromJson(JSONObject(this as Map<*, *>).toString(), FinderConfig::class.java)
}

fun Json.toSelectionOverlayConfig(gson: Gson): SelectionOverlayConfiguration {
    return gson.fromJson(
        JSONObject(this as Map<*, *>).toString(),
        SelectionOverlayConfiguration::class.java
    )
}

fun Json.toCroppingViewConfig(gson: Gson): CroppingViewConfig {
    return gson.fromJson(JSONObject(this as Map<*, *>).toString(), CroppingViewConfig::class.java)
}

fun Json.toContourConfig(gson: Gson): ContourConfiguration {
    return gson.fromJson(JSONObject(this as Map<*, *>).toString(), ContourConfiguration::class.java)
}

fun Json.toBarcodeScannerConfig(gson: Gson): BarcodeScannerConfiguration {
    return gson.fromJson(
        JSONObject(this as Map<*, *>).toString(),
        BarcodeScannerConfiguration::class.java
    )
}

fun Json.toMedicalCertificateConfig(gson: Gson): MedicalCertificateScannerConfiguration {
    return gson.fromJson(
        JSONObject(this as Map<*, *>).toString(),
        MedicalCertificateScannerConfiguration::class.java
    )
}

fun Json.toDocumentScannerConfig(gson: Gson): DocumentScannerConfiguration {
    return gson.fromJson(
        JSONObject(this as Map<*, *>).toString(),
        DocumentScannerConfiguration::class.java
    )
}

fun Json.toPageProcessorSizeConfig(gson: Gson): PageStorageProcessor.Configuration.Size {
    return gson.fromJson(
        JSONObject(this as Map<*, *>).toString(),
        PageStorageProcessor.Configuration.Size::class.java
    )
}

fun Page.toJson(gson: Gson): JsonObject {
    return gson.toJsonTree(this) as JsonObject
}

fun Json.toSize(gson: Gson): Page.Size {
    return gson.fromJson(JSONObject(this as Map<*, *>).toString(), Page.Size::class.java)
}

// Using Android's base64 libraries. This can be replaced with any base64 library.
class ByteArrayToBase64TypeAdapter : TypeAdapter<ByteArray>() {
    override fun write(out: JsonWriter?, value: ByteArray?) {
        out?.value(value?.let { Base64.encodeToString(it, Base64.NO_WRAP) } ?: "")
    }

    override fun read(inner: JsonReader?): ByteArray {
        return ByteArray(0)// we do not have any byte arrays going in sdk
    }
}

class PageSizeAdapter : TypeAdapter<Page.Size>() {
    override fun write(out: JsonWriter?, value: Page.Size) {
        out?.beginObject()
        out?.name("width")?.value(value?.width?.toDouble())
        out?.name("height")?.value(value?.height?.toDouble())
        out?.endObject()
    }

    override fun read(inner: JsonReader?): Page.Size {
        inner?.beginObject()
        var width = Int.MAX_VALUE
        var height = Int.MAX_VALUE
        while (inner?.hasNext() == true) {
            when (inner.nextName()) {
                "width" -> {
                    width = inner.nextDouble().toInt()
                }

                "height" -> {
                    height = inner.nextDouble().toInt()
                }
            }
        }
        inner?.endObject()
        return Page.Size(width, height)
    }
}

class PageProcessorSizeAdapter : TypeAdapter<PageStorageProcessor.Configuration.Size>() {
    override fun write(out: JsonWriter?, value: PageStorageProcessor.Configuration.Size?) {
        out?.beginObject()
        out?.name("width")?.value(value?.width?.toDouble())
        out?.name("height")?.value(value?.height?.toDouble())
        out?.endObject()
    }

    override fun read(inner: JsonReader?): PageStorageProcessor.Configuration.Size {
        inner?.beginObject()
        var width = Double.MAX_VALUE
        var height = Double.MAX_VALUE
        while (inner?.hasNext() == true) {
            when (inner.nextName()) {
                "width" -> {
                    width = inner.nextDouble()
                }

                "height" -> {
                    height = inner.nextDouble()
                }
            }
        }
        inner?.endObject()
        return PageStorageProcessor.Configuration.Size(width.toInt(), height.toInt())

    }
}

fun <T> Map<String, Any>.checkParam(name: String, value: (T) -> Unit = {}): T? {
    return this.get(name)?.also {
        value(it as T)
    } as? T
}
