//
//  BarcodeScanner.swift
//  scanbot_sdk
//
//  Created by Serhii Chaban on 27.06.22.
//

import Foundation
import ScanbotSDK

class BarcodeScanner: NSObject, ScannerDelegate, CallErrorHandler {
    private var viewController: SBSDKBarcodeScannerViewController?
    private var cameraMethodChannel: FlutterMethodChannel?
    private let detectorChannel: FlutterMethodChannel?
    private let finderChannel: FlutterMethodChannel?
    private var configuration: [String: Any?]
    private var shouldPauseDetectionOnResultReturn: Bool = false
    internal var shouldRunDetection: Bool = true
    
    init(
        cameraMethodChannel: FlutterMethodChannel?,
        detectorChannel: FlutterMethodChannel?,
        finderChannel: FlutterMethodChannel?,
        configuration: [String: Any?]
    ) {
        self.cameraMethodChannel = cameraMethodChannel
        self.detectorChannel = detectorChannel
        self.finderChannel = finderChannel
        self.configuration = configuration
    }
    
    func createNativeView(view: UIView) {
        if !ScanbotSDK.isLicenseValid() {
            DispatchQueue.main.async { [weak self] in
                guard let callHandler = self?.detectorChannel else {
                    fatalError("callHandler cannot be nil")
                }
                LicenseHandler.sendLicenseErrorToDetector(callHandler)
            }
            return
        }
        viewController = SBSDKBarcodeScannerViewController()
        viewController?.delegate = self
        
        viewController?.attachFlutterView(view)
        _ = viewController?.view
        
        setBarcodeScannerConfiguration(configuration: configuration)
        cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnCameraOpenMethod, arguments: "")
    }
    
    func attach(view: UIView) {
        createNativeView(view: view)
        self.detectorChannel?.setMethodCallHandler({ call, result in
            self.handleDetectorCall(call: call, result: result)
        })
    }
    
    func setFlash(enabled: Bool) {
        viewController?.isFlashLightEnabled = enabled
    }
    
    func isFlashAvailable() -> Bool {
        return viewController?.isFlashLightAvailable ?? false
    }
    
    func didChangeViewFinderRect(rect: CGRect) {
        finderChannel?.invokeMethod(MethodChannelConstants.kFinderSizeChanged, arguments: rect.rectDictionary)
    }
    
    private func sendDetectedBarcodesResult(_ codes: [SBSDKBarcodeScannerResult],
                                            _ barcodeImageURI: String?,
                                            completion: @escaping () -> Void) {
        let resultDict: [String: Any?] = ["barcodeItems": codes.map({ $0.encodedDictionary }),
                                          "barcodeImageURI": barcodeImageURI]
        DispatchQueue.main.async { [weak self] in
            guard let callHandler = self?.detectorChannel else {
                fatalError("callHandler cannot be nil")
            }
            callHandler.invokeMethod(MethodChannelConstants.kOnBarcodeDetected, arguments: resultDict.encodedString ) { _ in
                completion()
            }
    
        }
    }
    
    private func sendTappedBarcodeResult(_ code: SBSDKBarcodeScannerResult, completion: (() -> Void)? = nil) {
        
        DispatchQueue.main.async { [weak self] in 
            guard let callHandler = self?.cameraMethodChannel else {
                fatalError("callHandler cannot be nil")
            }
            callHandler.invokeMethod(MethodChannelConstants.kOnBarcodeSelected, arguments: code.encodedDictionary.encodedString) { _ in 
                if completion != nil {
                    completion!()
                }
                
            }
        }
        
    }
    
    func setBarcodeScannerConfiguration(configuration: [String: Any?]?) {
        viewController?.viewFinderRectDidChangeHandler = { (finderRect: CGRect) -> Void in
            self.didChangeViewFinderRect(rect: finderRect)
        }
        let barcodeConfiguration = BarcodeScannerConfiguration(configuration)
        
        if let flashEnabled = barcodeConfiguration.flashEnabled {
            viewController?.isFlashLightEnabled = flashEnabled
        }
        
        if let zoomConfiguration = barcodeConfiguration.zoomConfiguration {
            viewController?.zoomConfiguration = zoomConfiguration
        }
        
        if let viewFinderConfiguration = barcodeConfiguration.viewFinderConfiguration {
            viewController?.viewFinderConfiguration = viewFinderConfiguration
        }
        
        if let scannerConfiguration = barcodeConfiguration.scannerConfiguration {
            
            if let acceptedDocumentTypes = scannerConfiguration.acceptedDocumentTypes {
                viewController?.acceptedDocumentTypes = acceptedDocumentTypes
            }
            
            if let acceptedBarcodeTypes = scannerConfiguration.acceptedBarcodeTypes {
                viewController?.acceptedBarcodeTypes = acceptedBarcodeTypes
            }
            
            if let engineMode = scannerConfiguration.engineMode {
                viewController?.engineMode = engineMode
            }
            
            if let additionalDetectionParameters = scannerConfiguration.additionalDetectionParameters {
                viewController?.additionalDetectionParameters = additionalDetectionParameters
            }
            
            if let enableGS1Decoding = scannerConfiguration.enableGS1Decoding {
                viewController?.additionalDetectionParameters.enableGS1Decoding = enableGS1Decoding
            }
            
            if let minimumTextLength = scannerConfiguration.minimumTextLength {
                viewController?.additionalDetectionParameters.minimumTextLength = minimumTextLength
            }
            
            if let maximumTextLength = scannerConfiguration.maximumTextLength {
                viewController?.additionalDetectionParameters.maximumTextLength = maximumTextLength
            }
            
            if let minimum1DBarcodesQuietZone = scannerConfiguration.minimum1DBarcodesQuietZone {
                viewController?.additionalDetectionParameters.minimum1DBarcodesQuietZone = minimum1DBarcodesQuietZone
            }
            
            if let msiPlesseyChecksumAlgorithm = scannerConfiguration.msiPlesseyChecksumAlgorithm {
                viewController?.additionalDetectionParameters.msiPlesseyChecksumAlgorithm = msiPlesseyChecksumAlgorithm
            }
            
            if let stripCheckDigits = scannerConfiguration.stripCheckDigits {
                viewController?.additionalDetectionParameters.stripCheckDigits = stripCheckDigits
            }
            
            if let barcodeImageGenerationType = scannerConfiguration.barcodeImageGenerationType {
                viewController?.barcodeImageGenerationType = barcodeImageGenerationType
            }
        }
        
        if let selectionOverlayConfiguration = barcodeConfiguration.selectionOverlayConfiguration {
            viewController?.selectionOverlayEnabled = selectionOverlayConfiguration.selectionOverlayEnabled
            viewController?.selectionOverlayTextFormat = selectionOverlayConfiguration.selectionOverlayTextFormat
            viewController?.selectionPolygonColor = selectionOverlayConfiguration.selectionPolygonColor
            viewController?.selectionTextColor = selectionOverlayConfiguration.selectionTextColor
            viewController?.selectionTextContainerColor = selectionOverlayConfiguration.selectionTextContainerColor
            viewController?.automaticSelectionEnabled = selectionOverlayConfiguration.automaticSelectionEnabled
        }
    }
    
    func stopPreview() {
        viewController?.freezeCamera()
    }
    
    func resumePreview() {
        viewController?.unfreezeCamera()
    }
    
    func handleDetectorCall(call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case "enable": shouldRunDetection = true
        case "disable": shouldRunDetection = false
        default:
            self.returnUnknownCallError(call: call, result: result)
            return
        }
        resurnSuccess(result: result) // always return some result to the flutter
    }
    
    func resurnSuccess(result: @escaping FlutterResult) {
        result("success")
    }
}

extension BarcodeScanner: SBSDKBarcodeScannerViewControllerDelegate {
    
    func barcodeScannerController(_ controller: SBSDKBarcodeScannerViewController,
                                  didDetectBarcodes codes: [SBSDKBarcodeScannerResult]) {
        shouldPauseDetectionOnResultReturn = true
        var uri: URL?
        if !codes.isEmpty {
            if let image = codes[0].sourceImage {
                cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnShowHeavyProcessing, arguments: true)
                guard let uuid = SBSDKUIPageFileStorage.default().add(image) else { return  }
                uri = SBSDKUIPageFileStorage.default().imageURL(withPageFileID: uuid, pageFileType: .original)
                cameraMethodChannel?.invokeMethod(MethodChannelConstants.kOnShowHeavyProcessing, arguments: false)
            }
        }
        sendDetectedBarcodesResult(codes, uri?.standardizedFileURL.absoluteString) { [weak self] in
            self?.shouldPauseDetectionOnResultReturn = false
        }
    }
    
    func barcodeScannerController(_ controller: SBSDKBarcodeScannerViewController, didTapOnBarcode code: SBSDKBarcodeScannerResult) {
        sendTappedBarcodeResult(code, completion: nil)
    }
    
    func barcodeScannerControllerShouldDetectBarcodes(_ controller: SBSDKBarcodeScannerViewController) -> Bool {
        return shouldRunDetection && !shouldPauseDetectionOnResultReturn
    }
}
