import 'dart:core';
import 'package:collection/collection.dart';

import 'package:json_annotation/json_annotation.dart';
import 'generic_document_wrappers.dart';

import 'common.dart';
import 'generic_document.dart';
export 'generic_document.dart';

part 'common_generic_document.g.dart';

/// Extension function to GenericDocument to make it easier to access fields.
extension GenericDocumentExtension on GenericDocument {
  /// Returns a field given its local or common type name. Returns null if not found.
  Field? fieldByTypeName(String name) {
    Field? field =
        fields?.firstWhereOrNull((element) => element.type.name == name);

    if (field != null) {
      return field;
    }

    return fields
        ?.firstWhereOrNull((element) => element.type.commonType?.name == name);
  }

  /// Returns all fields given a local or common field type name.
  List<Field> fieldsByTypeName(String name) {
    return fields?.where((element) => element.type.name == name).toList() ?? [];
  }

  /// Returns a sub-document given its document type name. Returns null if not found.
  GenericDocument? childByDocumentType(String name) {
    return children?.firstWhereOrNull((element) => element.type?.name == name);
  }
}

class FieldWrapper {
  /// Field type
  late FieldType type;

  FieldWrapper(Field field) {
    type = field.type;
  }
}

/// Text field wrapper
class TextFieldWrapper extends FieldWrapper {
  /// Field text value
  OcrResult? value;

  TextFieldWrapper(Field field) : super(field) {
    value = field.value;
  }
}

/// Text field wrapper
class ValidatedTextFieldWrapper extends TextFieldWrapper {
  /// Field text value
  bool validated = false;

  ValidatedTextFieldWrapper(Field field) : super(field) {
    validated = field.validationStatus! == ValidationStatus.Valid;
  }
}

class GenericDocumentWrapper {
  late GenericDocument document;

  GenericDocumentWrapper(this.document);

  DocumentType get documentType => DocumentType.Unknown;

  factory GenericDocumentWrapper.fromGenericDocument(
          GenericDocument document) =>
      wrapToDocumentType(document);
}

@JsonSerializable(
  includeIfNull: false,
  explicitToJson: false,
  createToJson: false,
)
class GenericDocumentResults {
  List<GenericDocument> documents = [];

  GenericDocumentResults(this.documents);

  factory GenericDocumentResults.fromJson(Map<String, dynamic> json) =>
      _$GenericDocumentResultsFromJson(json);
}

class GenericDocumentWrappedResults extends Result {
  List<GenericDocumentWrapper> documents = [];

  GenericDocumentWrappedResults(this.documents);

  GenericDocumentWrappedResults.error(String message) {
    operationResult = OperationResult.ERROR;
    this.message = message;
  }

  GenericDocumentWrappedResults.canceled() {
    operationResult = OperationResult.CANCELED;
  }

  factory GenericDocumentWrappedResults.fromGenericDocumentResults(
      GenericDocumentResults results) {
    List<GenericDocumentWrapper> documents = [];
    results.documents.forEach((element) {
      documents.add(GenericDocumentWrapper.fromGenericDocument(element));
    });
    return GenericDocumentWrappedResults(documents);
  }
}
